// Swap key/value pairs in object
var swapObject = function(obj) {
	var reverseObj = {};
	for(var key in obj) {
		reverseObj[obj[key]] = key;
	}
	return reverseObj;
}

var commandLength = {  // #bits in each command
	'REG_WRITE_IMM': 22,		
	'SET_DELAY': 22,			// Compiles in to REG_WRITE_IMM statement
	'REG_WRITE_MASK_IMM': 30,	
	'REG_WRITE_VOUT_IMM': 24,
	'REG_WRITE_VCTRL_IMM': 32,	
	'REG_WRITE_MASK_SREG': 24,		
	'SREG_READ_REG': 16,				
	'WAIT': 24,
	'JUMP': 24,					// Compiles in to WAIT statement
	'DELAY_IMM': 12,			
	'DELAY_SREG': 6,					
	'TRIG_SET': 32,				
	'TRIG_MASK': 32,			
	'END': 4
}

var regulatorIdMap = {
	'BUCK1': 0,
	'BUCK2': 1,
	'BUCK3': 2,
	'BUCK4': 3,
	'BUCK5': 4,
	'LDO1': 8,
	'LDO2': 9,
	'LDO3': 10,
	'LDO4': 11
}
var reverseRegulatorIdMap = swapObject(regulatorIdMap);

var registerIdMap = {
	'R0': '00',
	'R1': '01',
	'R2': '10',
	'R3': '11'
}

var opcodeMap = {
	'REG_WRITE_IMM': 		'0001',
	'SET_DELAY': 			'0001',
	'REG_WRITE_MASK_IMM':	'0010',
	'REG_WRITE_VOUT_IMM':	'0011',
	'REG_WRITE_VCTRL_IMM':	'0100',
	'REG_WRITE_MASK_SREG': 	'0101',
	'SREG_READ_REG': 		'0110',
	'WAIT': 				'0111',
	'JUMP':					'0111',
	'DELAY_IMM': 			'1000',
	'DELAY_SREG':			'1001',
	'TRIG_SET': 			'1010',
	'TRIG_MASK':			'1011',
	'END': 					'1100'		
}


var voutSelMap = {
	'VSET1':	'00',
	'VSET2':	'01',
	'ACTIVE':	'10',
	'INACTIVE':	'11',
	'0':		'00',	// Numbers 0-3 can be used instead of the labels
	'1':		'01',
	'2':		'10',
	'3':		'11'
}

var reverseVoutSelMap = swapObject(voutSelMap);

var conditionSelectMap = {
	'GPIO1':			'000000',
	'GPIO2':			'000001',
	'GPIO3':			'000010',
	'GPIO4': 			'000011',
	'GPIO5':			'000100',
	'GPIO6':			'000101',
	'GPIO7':			'000110',
	'GPIO8':			'000111',
	'GPIO9':			'001000',
	'GPIO10':			'001001',
	'GPIO11':			'001010',
	'BUCK1_PG':			'001011',
	'BUCK2_PG':			'001100',
	'BUCK3_PG':			'001101',
	'BUCK4_PG':			'001110',
	'BUCK5_PG':			'001111',
	'LDO1_PG':			'010000',
	'LDO2_PG':			'010001',
	'LDO3_PG':			'010010',
	'LDO4_PG':			'010011',
	'PGOOD':			'010100',
	'TWARN_EVENT':		'010101',
	'INTERRUPT_PIN':	'010110',
	//
	'I2C_0':			'100000',
	'I2C_1':			'100001',
	'I2C_2':			'100010',
	'I2C_3':			'100011',
	'I2C_4':			'100100',
	'I2C_5':			'100101',
	'I2C_6':			'100110',
	'I2C_7':			'100111',
	'SREG0_0':			'101000',
	'SREG0_1':			'101001',
	'SREG0_2':			'101010',
	'SREG0_3':			'101011',
	'SREG0_4':			'101100',
	'SREG0_5':			'101101',
	'SREG0_6':			'101110',
	'SREG0_7':			'101111',
	'LP_STANDBY_SEL':	'110000',
	'0':				'111110',
	'1':				'111111'
}
var reverseConditionSelectMap = swapObject(conditionSelectMap);

var triggerSelectMap = {
	'IMMEDIATE_SHUTDOWN': 		'000000',
	'MCU_POWER_ERROR':			'000001',
	'ORDERLY_SHUTDOWN':			'000010',
	//'N/A':					'000011',
	'FORCE_STANDBY':			'000100',
	'FORCE_LP_STANDBY':			'000101',
	'ESM_MCU_ERROR':			'000110',
	'WD_ERROR':					'000111',
	'SOC_POWER_ERROR':			'001000',
	'ESM_SOC_ERROR':			'001001',
	'A':						'001010',
	'WKUP1':					'001011',
	'SU_ACTIVE':				'001100',
	'B':						'001101',
	'WKUP2':					'001110',
	'SU_MCU_ONLY':				'001111',
	'C':						'010000',
	'D':						'010001',
	'SU_STANDBY':				'010010',
	'SU_X':						'010011',
	'WAIT_TIMEOUT':				'010100',
	'GPIO1':					'010101',
	'GPIO2':					'010110',
	'GPIO3':					'010111',
	'GPIO4':					'011000',
	'GPIO5':					'011001',
	'GPIO6':					'011010',
	'GPIO7':					'011011',
	'GPIO8':					'011100',
	'GPIO9':					'011101',
	'GPIO10':					'011110',
	'GPIO11':					'011111',
	'I2C_0':					'100000',
	'I2C_1':					'100001',
	'I2C_2':					'100010',
	'I2C_3':					'100011',
	'I2C_4':					'100100',
	'I2C_5':					'100101',
	'I2C_6':					'100110',
	'I2C_7':					'100111',
	'SREG0_0':					'101000',
	'SREG0_1':					'101001',
	'SREG0_2':					'101010',
	'SREG0_3':					'101011',
	'SREG0_4':					'101100',
	'SREG0_5':					'101101',
	'SREG0_6':					'101110',
	'SREG0_7':					'101111',
	'LP_STANDBY_SEL':			'110000',
	'0':						'111110',
	'1':						'111111'
}
var reverseTriggerSelectMap = swapObject(triggerSelectMap);

var triggerTypeMap = {
	'LOW':  '00',
	'HIGH': '01',
	'RISE': '10',
	'FALL': '11'
}
var reverseTriggerTypeMap = swapObject(triggerTypeMap);

var externalStateMap = {
	'RESET':				0,
	'NVM_LDO_SU':			1,
	'NVM_READ':				2,
	'CHANGE_BG_DELAY':		3,
	'NARROW_LIMITS_DELAY':	4,
	'BOOT_BIST':			5,
	'MISSION_STATES':		6,
	'SAFE_RECOVERY':		7,
	'RUNTIME_BIST':			8,
	'USER_EE_PROG_SU':		9,
	'USER_EE_PROG':			10,
	'LP_STANDBY':           11,
	'OFF':					15
}

var reverseExternalStateMap = swapObject(externalStateMap);
	
// File contents is a string containing the input assembly
// debug is a boolean, when true it will include debug output
var assemble = function (fileContents, debug) {
    
    var programTxt = [];			// Text of each instruction
    var programBin = [];			// Binary of each instruction
	var programAddrList = [];		// List of addresses for each instruction
	
	// This is used for the self-test at the end, if debug=true
	var expectedProgramBin = [
	'00000011100100010000110011111010', 	// TRIG_SET DEST=seq1 ID=15 SEL=ESM_MCU_ERROR TYPE=RISE IMM=0 EXT=0
	'00000000001111100010101011001010',		// TRIG_SET DEST=SAFE_RECOVERY ID=12 SEL=GPIO1         TYPE=LOW IMM=1 EXT=1
	'10101010101111',						// LITERAL 4,0b1111 10,0x2aa ;invalid command, for testing only
	'01010101010101010101010101011011',		// TRIG_MASK 0x5555555
	'1100',									// END
	'0000011000100010110001',				// SET_DELAY 3.2us
	'000010000100011100000111',				// WAIT COND=LP_STANDBY_SEL TYPE=HIGH TIMEOUT=13us dest=seq2
	'0000010100100010110001',				// SET_DELAY 1.6us
	'1100',									// END
	'1011001100110110',						// REG_READ R2 0x333
	'1110100010000110',						// REG_READ R3 0x288
	'00011000000001000000011010100100',		// REG_WRITE_VCTRL_IMM REGULATOR=LDO3 DELAY=10us MASK=0x1 VCTRL=0x1 DELAY_MODE=ALWAYS
	'00110100011111000110100000100100',		// REG_WRITE_VCTRL_IMM BUCK3 0b11010 0x1F 20us MATCH_EN
	'00011011111111111111110110000100',		// REG_WRITE_VCTRL_IMM REGULATOR=LDO1 DELAY=10us MASK=0xFF VCTRL=0xFF DELAY_MODE=MATCH_ALL
	'0010001000000111010001',				// REG_WRITE_IMM DATA=0x22 ADDR=0x1D
	'0010001000000111010001',
	'0010001000000111010001',
	'0100010011001100110001',				// REG_WRITE_IMM ADDR=0x433 DATA=0x44
	'010101010000010011001100110010',
	'1100',									// END
	'110010011100101000010011',				// REG_WRITE_VOUT_IMM BUCK2 2 1.095V 80us
	'000110111111110000000011',				// REG_WRITE_VOUT_IMM BUCK1 VSET1 3.34V 10us
	'011001011100000010000011',				// REG_WRITE_VOUT_IMM REGULATOR=LDO1 VOUT=500mV SEL=0 DELAY=80us BYPASS=0
	'000010101110000010100011',				// REG_WRITE_VOUT_IMM REGULATOR=LDO3 VOUT=800mV SEL=0 DELAY=3us BYPASS=1
	'000001101000',
	'000000111000',
	'100101101000',
	'000010101000',
	'001001',
	'111001',
	'000001000000001111110111',				// JUMP dest=seq2 ext=0
	'0010001000000111010001',				// REG_WRITE_IMM DATA=0x22 ADDR=0x1D
	'1100'									// END
	]
	
	
	
	var symbolsAddress = {}; // symbol to address mapping
	var reverseSymbols = {}; // address to symbol mapping
	var symbolsInstructionCount = {};
    var programAddress = 0;
	
	var error = false;
	var defaultDelayStep = -1;
	var currentDelayStep = -1;
	var inInitSeq = false;			// Used to set default delay value if there is a SET_DELAY command in the initialization sequence
	var parameters = {};
	var startSeqFound = false;
	var instructionCount = 0;
	
    var lines = fileContents.split("\n");
	
    // Skip all static information before the sequence.
	// Only look for PFSM_DELAY_STEP to determine the default delay step size from NVM
    for(var lineNum=0; lineNum < lines.length; lineNum++) {
        if(lines[lineNum].indexOf("pfsm_start") === 0) {
            startSeqFound = true;
            break;
        } else if(lines[lineNum].indexOf("PFSM_DELAY_STEP") !== -1) {
			delayValue = parseInt(lines[lineNum].split("=")[1]);
			defaultDelayStep = 50 * Math.pow(2, delayValue);
			currentDelayStep = defaultDelayStep;
		}
    }
    if(!startSeqFound) {
        console.log("ERROR: Did not find 'pfsm_start' label.");
        return "";
    }
	
	inInitSeq = true;

    // First pass of program, to find symbol labels.  Store valid instructions in to a program text array for future passes.
    for(; lineNum < lines.length; lineNum++) {
        
		line = lines[lineNum];
		
		if(line.indexOf(";") !== -1) {
			line = line.substring(0, line.indexOf(";")); // Remove comments
		}
		
		line = line.trim();
		
        if(line.indexOf(":") !== -1) {
            symbolName = line.split(":")[0];
			if(symbolsAddress.hasOwnProperty(symbolName)) {
				console.log("ERROR: duplicate label '" + symbolName + "'");
				error = true;
			}
            symbolsAddress[symbolName] = programAddress;
			reverseSymbols[programAddress] = symbolName;
			symbolsInstructionCount[symbolName] = instructionCount;
			line = line.substring(line.indexOf(":")+1).trim();
        }		
		
		if(line === "") {
            continue;
		} else {
            fields = line.split(" ");
			fields = fields.filter(function (el) {
				return el != "";	// delete empty elements due to multiple consecutive spaces
			});
            command = fields[0];
			programAddrList.push(programAddress)
			
			if(command === "LITERAL") {
				var binary = parseLiteralCommand(fields, line);
				if(binary === undefined) {
					error = true;
					continue;
				}
				
				programAddress += binary.length;
			} else if(commandLength[command] === undefined) {
                console.log("Invalid command: " + command);
                return "";
            } else {
				programAddress += commandLength[command];
			}
            
            programTxt.push(line);
			instructionCount += 1;			
        }
    }
    
    for(var i=0; i<programTxt.length; i++) {
        var fields = programTxt[i].split(" ");
		fields = fields.filter(function (el) {
			return el != "";	// delete empty elements due to multiple consecutive spaces
		});
        var command = fields[0];
        
        if(command === "END") {
            programBin[i] = opcodeMap[command];		
			inInitSeq = false;
			currentDelayStep = defaultDelayStep;
        } else if(command === "REG_WRITE_IMM") {
			
			parameters = parseParameters(["ADDR", "DATA"], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}

			var addr = parseInt(parameters['ADDR']).toString(2);
			var data = parseInt(parameters['DATA']).toString(2);
			
			addr = calculateAddress(addr, programTxt[i]);
			if(addr === -1) {
				error = true;
				continue;
			}
			
			if(data === "NaN") {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid data value, expecting a number");
				error = true;
				continue;
			}
			
			if(data.length > 8) {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid data value, must be 8 bits or less");
				error = true;
				continue;
			}
			
			data = zeroPad(data, 8);
						
			programBin[i] = data + addr + opcodeMap[command];
			
		} else if(command === "SET_DELAY") {
			
			parameters = parseParameters(["STEP"], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			var targetStep = convertTimeToNs(parameters["STEP"]);
			if(targetStep === -1) {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid delay step '" + parameters["STEP"] + "'\nExpecting a number with units of ns, us, ms, or s");
			}
			var delayFound = false;
			
			// Check if time step is too high or too low
			if(targetStep > 107.4e9) {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid delay step '" + parameters["STEP"] + "'\nMaximum step size is 107.4 seconds");
				error = true;
				break;
			} else if(targetStep < 50) {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid delay step '" + parameters["STEP"] + "'\nMinimum step size is 50ns");
				error = true;
				break;
			}
			
			var delayValue = 0;
			for(var j=0; j<32; j++) {
				var delayStep = 50 * Math.pow(2,j);
				// If specified delay is within 10% of a time step, round to that nearest step
				if(Math.abs(delayStep-targetStep) / targetStep < 0.1) {
					currentDelayStep = delayStep;
					if(inInitSeq) {
						defaultDelayStep = delayStep;
					}
					delayValue = j;
					delayFound = true;
					break;
				}
				// If no time step is within 10% of the specified delay, throw an error
				if(delayStep / targetStep > 1.1) {
					console.log("\nInstruction: " + programTxt[i] + "\nInvalid delay step: " + fields[1] + ".");
					console.log("Choose step size of " + convertNsToTime(50*Math.pow(2,j-1)) + " or " + convertNsToTime(50*Math.pow(2,j)));
					error = true;
					break;
				}
			}
			
			if(!delayFound) {
				continue;
			}
			
			if(fields[1] !== convertNsToTime(currentDelayStep)) {
				console.log("INFO: specified delay step of " + fields[1] + " is converted to delay step of " + convertNsToTime(currentDelayStep));
			}
			
			delayBin = delayValue.toString(2);
			
			// Command is 8-bit 0-padded delay value, plus 10-bit address 0x8B, plus 4-bit command 0x1
			programBin[i] = zeroPad(delayBin, 8) + "0010001011" + opcodeMap[command];
			
		} else if(command === "REG_WRITE_MASK_IMM") {
			
			parameters = parseParameters(['ADDR', 'DATA', 'MASK'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			var addr = parseInt(parameters['ADDR']).toString(2);
			var data = parseInt(parameters['DATA']).toString(2);
			var mask = parseInt(parameters['MASK']).toString(2);
			
			var addr = calculateAddress(addr, programTxt[i]);
			if(addr === -1) {
				error = true;
				continue;
			}
			
			if(data === "NaN") {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid data value, expecting a number");
				error = true;
				continue;
			}
			if(mask === "NaN") {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid mask value, expecting a number");
				error = true;
				continue;
			}
			
			if(data.length > 8) {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid data value, must be 8 bits or less");
				error = true;
				continue;
			}
			data = zeroPad(data, 8);
			
			if(mask.length > 8) {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid mask value, must be 8 bits or less");
				error = true;
				continue;
			}
			mask = zeroPad(mask, 8);
			
			programBin[i] = mask + data + addr + opcodeMap[command];
			
		} else if(command === "REG_WRITE_VOUT_IMM") {
			
			parameters = parseParameters(['REGULATOR', 'SEL', 'VOUT', 'DELAY', 'BYPASS (optional)'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			if(!regulatorIdMap.hasOwnProperty(parameters['REGULATOR'].toUpperCase())) {
				console.log("\Instruction: " + programTxt[i] + "\nERROR: Invalid regulator '" + parameters['REGULATOR'] + "', expecting BUCK1-BUCK5, or LDO1-LDO5");
				error = true;
				continue;
			}
			var regulator = regulatorIdMap[parameters['REGULATOR'].toUpperCase()].toString(2);
			regulator = zeroPad(regulator, 4);
			
			if(!voutSelMap.hasOwnProperty(parameters['SEL'].toUpperCase())) {
				console.log("Instruction: " + programTxt[i] + "\nEROR: Invalid VOUT select type '" + parameters['SEL'] + "', expecting VSET1, VSET2, ACTIVE, or INACTIVE");
				error = true;
				continue;
			}
			var sel = voutSelMap[parameters['SEL'].toUpperCase()];
			
			var voutStr = parameters['VOUT'];
			var vout = 0;	// Vout value in mV
			
			// Validate VOUT is a number ending in mV or V
			if(voutStr.substring(voutStr.length-2).toUpperCase() === "MV") {
				vout = convertStringToInt(voutStr.substring(0, voutStr.length-2));
				if(isNaN(vout)) {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "', expecting a number with units of mV or V");
					error = true;
					continue;
				}
			} else if(voutStr.substring(voutStr.length-1).toUpperCase() === "V") {
				vout = convertStringToInt(voutStr.substring(0, voutStr.length-1))*1000;
				if(isNaN(vout)) {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "', expecting a number with units of mV or V");
					error = true;
					continue;
				}
			} else {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "', expecting a number with units of mV or V");
				error = true;
				continue;
			}
			
			var vset = 0;
			
			// Validate VOUT is an acceptable voltage value, and calculate the VSET value for that VOUT
			if(parameters['REGULATOR'].toUpperCase().indexOf("BUCK") !== -1) {
				if(vout < 300) {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "', must be between 0.3V and 3.34V for buck output voltage");
					error = true;
					continue;
				} else if(vout < 600) {
					vset = (vout-300) / 20;
					if(Math.round(vset) !== vset) {
						console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + (Math.floor(vset)*20 + 300) + "mV and " + (Math.ceil(vset)*20 + 300) + "mV");
						error = true;
						continue;
					}
				} else if(vout < 1100) {
					vset = ((vout-600) / 5) + 15;
					if(Math.round(vset) !== vset) {
						console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-15)*5 + 600) + "mV and " + ((Math.ceil(vset)-15)*5 + 600) + "mV");
						error = true;
						continue;
					}
				} else if(vout < 1660) {
					vset = ((vout-1100) / 10) + 0x73;
					if(Math.round(vset) !== vset) {
						console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-0x73)*10 + 1100) + "mV and " + ((Math.ceil(vset)-0x73)*10 + 1100) + "mV");
						error = true;
						continue;
					}
				} else if(vout <= 3340) {
					vset = ((vout-1660) / 20) + 0xAB;
					if(Math.round(vset) !== vset) {
						console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-0xAB)*20 + 1660) + "mV and " + ((Math.ceil(vset)-0xAB)*20 + 1660) + "mV");
						error = true;
						continue;
					}
				} else {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "', must be between 0.3V and 3.34V");
					error = true;
					continue;				
				}
				
				if(parameters['BYPASS'] !== undefined) {
					console.log("\nInstruction: " + programTxt[i] + "\nINFO: Bypass mode is not valid for BUCK regulators.  Ignoring bypass bit");
				}
				
			} else { // LDO
				if (vout < 600) {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "', must be between 0.6V and 3.3V for LDO output voltage");
					error = true;
					continue;
				} else if(vout <= 3300) {
					var ldo = parameters['REGULATOR'].toUpperCase();
					if(ldo === "LDO4") {
						vset = ((vout-600) / 25) + 0x8;
						if(Math.round(vset) !== vset) {
							console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-0x8)*25 + 600) + "mV and " + ((Math.ceil(vset)-0x8)*25 + 600) + "mV");
							error = true;
							continue;
						}	
					} else { // LDO1-3
						vset = ((vout-600) / 50) + 0x4;
						if(Math.round(vset) !== vset) {
							console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-0x4)*50 + 600) + "mV and " + ((Math.ceil(vset)-0x4)*50 + 600) + "mV");
							error = true;
							continue;
						}					
						vset = vset << 1;
					}

				} else {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VOUT '" + voutStr + "', must be between 0.6V and 3.3V for LDO output voltage");
					error = true;
					continue;	
				}
				
				if(parameters['BYPASS'] !== undefined && parameters['BYPASS'] !== "0" && parameters['BYPASS'] !== "1") {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Bypass mode must be 1 or 0 when writing an LDO output voltage");
					error = true;
					continue;
				} else {
					if(parameters['BYPASS'] === "1") {
						vset |= 0x80;
					}
				}
			}
			
			vset = vset.toString(2);
			vset = zeroPad(vset, 8);
			
			var delayVal = convertDelayTimeToValue(parameters['DELAY'], currentDelayStep, programTxt[i], 6);
			if(delayVal === -1) {
				error = true;
				continue;
			}
			
			delayVal = delayVal.toString(2);
			delayVal =zeroPad(delayVal, 6);
			
			programBin[i] = delayVal + vset + sel + regulator + opcodeMap[command];
			
		} else if(command === "REG_WRITE_VCTRL_IMM") {
			
			parameters = parseParameters(['REGULATOR', 'VCTRL', 'MASK', 'DELAY', 'DELAY_MODE'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			if(!regulatorIdMap.hasOwnProperty(parameters['REGULATOR'].toUpperCase())) {
				console.log("\Instruction: " + programTxt[i] + "\nERROR: Invalid regulator '" + parameters['REGULATOR'] + "', expecting BUCK1-BUCK5, or LDO1-LDO5");
				error = true;
				continue;
			}
			var regulator = regulatorIdMap[parameters['REGULATOR'].toUpperCase()].toString(2);
			regulator = zeroPad(regulator, 4);
			
			var vctrl = convertStringToInt(parameters['VCTRL']).toString(2);
			if(vctrl === "NaN") {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VCTRL value '" + parameters['VCTRL'] + "', expecting a number");
				error = true;
				continue;
			}
			if(vctrl.length > 8) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid VCTRL value '" + parameters['VCTRL'] + "', must be 8 bits or less");
				error = true;
				continue;
			}
			vctrl = zeroPad(vctrl, 8);
			
			var mask = convertStringToInt(parameters['MASK']).toString(2);
			if(mask === "NaN") {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid mask value '" + parameters['MASK'] + "', expecting a number");
				error = true;
				continue;
			}
			if(mask.length > 8) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid MASK value '" + parameters['MASK'] + "', must be 8 bits or less");
				error = true;
				continue;
			}
			mask = zeroPad(mask, 8);
			
			var delayVal = convertDelayTimeToValue(parameters['DELAY'], currentDelayStep, programTxt[i], 6);
			if(delayVal === -1) {
				error = true;
				continue;
			}
			delayVal = delayVal.toString(2);
			delayVal = zeroPad(delayVal, 6);
			
			var delayMode;
			if (parameters['DELAY_MODE'].toUpperCase() == "MATCH_EN") {
				delayMode = '00';
			} else if(parameters['DELAY_MODE'].toUpperCase() == 'MATCH_ALL') {
				delayMode = '01';
			} else if(parameters['DELAY_MODE'].toUpperCase() == 'ALWAYS') {
				delayMode = '10';
			} else {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid delay type '" + parameters['DELAY_MODE'] + "'.  Valid options are:");
				console.log("  MATCH_EN: delay only if enable bit is different");
				console.log("  MATCH_ALL: delay if any bit is different");
				console.log("  ALWAYS: Always delay");
				error = true;
				continue;
			}
			
			// Do something with delayType
			programBin[i] = delayVal + mask + vctrl + delayMode + regulator + opcodeMap[command];			
			
		} else if(command === "REG_WRITE_MASK_SREG") {
			
			parameters = parseParameters(['REG', 'ADDR', 'MASK'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			var reg = parameters['REG'].toUpperCase();
			if(!registerIdMap.hasOwnProperty(reg)) {
				console.log("\Instruction: " + programTxt[i] + "\nERROR: Invalid register '" + parameters['REG'] + "', expecting R0-R3");
				error = true;
				continue;
			}
			reg = registerIdMap[reg];
			
			var addr = parseInt(parameters['ADDR']).toString(2);
			addr = calculateAddress(addr, programTxt[i]);
			if(addr === -1) {
				error = true;
				continue;
			}
			
			var mask = parseInt(parameters['MASK']).toString(2);
			if(mask === "NaN") {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid mask value '" + parameters['MASK'] + "', expecting a number");
				error = true;
				continue;
			}
			
			if(mask.length > 8) {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid mask value '" + parameters['MASK'] + "', must be 8 bits or less");
				error = true;
				continue;
			}
			mask = zeroPad(mask, 8);
			
			programBin[i] = mask + reg + addr + opcodeMap[command];
			
		} else if(command === "SREG_READ_REG") {
			
			parameters = parseParameters(['REG', 'ADDR'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			var reg = parameters['REG'].toUpperCase();
			if(!registerIdMap.hasOwnProperty(reg)) {
				console.log("\Instruction: " + programTxt[i] + "\nERROR: Invalid register '" + parameters['REG'] + "', expecting R0-R3");
				error = true;
				continue;
			}
			reg = registerIdMap[reg];
			
			var addr = parseInt(parameters['ADDR']).toString(2);
			if(addr === "NaN") {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid register address '" + parameters['ADDR'] + "', expecting a number");
				error = true;
				continue;
			}
						
			if(addr.length > 10) {
				console.log("\nInstruction: " + programTxt[i] + "\nInvalid register address '" + parameters['ADDR'] + "', must be 10 bits or less");
				error = true;
				continue;
			}
			addr = zeroPad(addr, 10);
			
			programBin[i] = reg + addr + opcodeMap[command];
			
			
		} else if(command === "WAIT") {
			
			// Note, either DEST or SKIP_COUNT must be present
			parameters = parseParameters(['COND', 'TYPE', 'TIMEOUT', 'DEST (optional)', 'SKIP_COUNT (optional)'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			if(parameters['DEST'] === undefined && parameters['SKIP_COUNT'] === undefined) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Must specify either DEST or SKIP_COUNT");
				error = true;
				continue;
			}
			if(parameters['DEST'] !== undefined && parameters['SKIP_COUNT'] !== undefined) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Cannot specify both DEST and SKIP_COUNT.  Specify one or the other");
				error = true;
				continue;
			}
			
			var condition = parameters['COND'].toUpperCase();
			if(!conditionSelectMap.hasOwnProperty(condition)) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid condition '" + condition + "'");
				error = true;
				continue;
			}
			conditionVal = zeroPad(conditionSelectMap[condition], 6);
			
			var conditionType = parameters['TYPE'].toUpperCase();
			if(!triggerTypeMap.hasOwnProperty(conditionType)) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid trigger type '" + conditionType + "'.  Must be LOW, HIGH, RISE, or FALL");
				error = true;
				continue;
			}
			var conditionTypeVal = triggerTypeMap[conditionType];

			var timeout = convertDelayTimeToValue(parameters['TIMEOUT'], currentDelayStep, programTxt[i], 6);
			if(timeout === -1) {
				error = true;
				continue;
			}
			timeoutVal = timeout.toString(2);
			timeoutVal = zeroPad(timeoutVal, 6);
			
			// TODO: Warning if destination jumps over an END
			if(parameters['DEST'] !== undefined) { // DEST provided instead of SKIP_COUNT
				var destination = parameters['DEST'];
				if(!symbolsAddress.hasOwnProperty(destination)) {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid destination '" + destination + "'");
					error = true;
					continue;
				}
			
				var skipCount = symbolsInstructionCount[destination] - i - 1;	// 'i' contains current instruction count
				if(skipCount < 0 || skipCount > 63) {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid destination '" + destination + "'.\n  Destination is " + skipCount + " instructions after the WAIT statement, must be between 0 and 63.");
					error = true;
					continue;
				}
				skipCount = zeroPad(skipCount.toString(2), 6);
			} else { // SKIP_COUNT provided instead of DEST
				var skipCount = parseInt(parameters['SKIP_COUNT']);
				if(skipCount < 0 || skipCount > 63) {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Skip count must be between 0 and 63");
					error = true;
					continue;
				}
				skipCount = skipCount.toString(2);
				skipCount = zeroPad(skipCount, 6);
			}
			
			programBin[i] = skipCount + timeoutVal + conditionTypeVal + conditionVal + opcodeMap[command];
			
		} else if(command === "JUMP") {
			
			parameters = parseParameters(['DEST'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			var destination = parameters['DEST'];
			if(!symbolsAddress.hasOwnProperty(destination)) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid destination '" + destination + "', does not exist");
				error = true;
				continue;
			}
			var skipCount = symbolsInstructionCount[destination] - i - 1;	// 'i' contains current instruction count
			if(skipCount < 0 || skipCount > 63) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid destination '" + destination + "'.\n  Destination is " + skipCount + " instructions after the JUMP statement, must be between 0 and 63.");
				error = true;
				continue;
			}
			skipCount = zeroPad(skipCount.toString(2), 6);
			
			var timeoutVal = '000000';
			var conditionTypeVal = triggerTypeMap['LOW'];
			var conditionVal = conditionSelectMap['1'];
			
			programBin[i] = skipCount + timeoutVal + conditionTypeVal + conditionVal + opcodeMap[command];			
			
		} else if(command === "DELAY_IMM") {
			
			parameters = parseParameters(['DELAY'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			var delayTxt = parameters['DELAY'];
			var delayVal = 0;
			if(delayTxt.substring(delayTxt.length-1).toUpperCase() !== "S") {
				// Assume there is no unit, try to parse as direct number
				delayVal = convertStringToInt(fields[1]);
				if(isNaN(delayVal) || delayVal < 0 || delayVal > 255) {
					console.log("\nInstruction: " + programTxt[i] + "\nInvalid delay time, expecting a number 0-255, or else a time with units of ns, us, ms, or s");
					error = true;
					continue
				}
			} else {
				delayVal = convertDelayTimeToValue(delayTxt, currentDelayStep, programTxt[i], 8);
				if(delayVal === -1) {
					error = true;
					continue;
				}
			}
			
			delayVal = delayVal.toString(2);
			delayVal = zeroPad(delayVal, 8);
			
			programBin[i] = delayVal + opcodeMap[command];
			
		} else if(command === "DELAY_SREG") {
			
			parameters = parseParameters(['REG'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			var reg = parameters['REG'].toUpperCase();
			if(!registerIdMap.hasOwnProperty(reg)) {
				console.log("\Instruction: " + programTxt[i] + "\nERROR: Invalid register '" + parameters['REG'] + "', expecting R0-R3");
				error = true;
				continue;
			}
			reg = registerIdMap[reg];
			
			programBin[i] = reg + opcodeMap[command];			
			
		} else if(command === "TRIG_SET") {

			parameters = parseParameters(['DEST', 'ID', 'SEL', 'TYPE', 'IMM', 'EXT'], fields, programTxt[i]);
			if(parameters === false) {
				error = true;
				continue;
			}
			
			var extVal = convertStringToInt(parameters['EXT']);
			if(isNaN(extVal) || extVal < 0 || extVal > 1) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid external value '" + parameters['EXT'] + "', must be 0 or 1");
				error = true;
				continue;
			}
			extVal = extVal.toString(2);
			
			var destination = parameters['DEST'];
			if(extVal == '0') {	// Jump to a label in the program
				if(!symbolsAddress.hasOwnProperty(destination)) {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid destination '" + destination + "'");
					error = true;
					continue;
				}
				var destinationVal = symbolsAddress[destination].toString(2);
			} else { 			// Jump to an external FFSM state
				if(!externalStateMap.hasOwnProperty(destination)) {
					console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid external destination '" + destination + "'.  Options are:");
					for(var property in externalStateMap) {
						if(externalStateMap.hasOwnProperty(property)) {
							console.log(property);
						}
					}
					error = true;
					continue;
				}
				var destinationVal = externalStateMap[destination].toString(2);				
			}
			if (destinationVal.length > 13) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Destination address 0x" + parseInt(destinationVal, 2).toString(16) + " exceeds 13 bits");
				error = true;
				continue;
			}
			destinationVal = zeroPad(destinationVal, 13);
			
			var trigIdVal = convertStringToInt(parameters['ID']);
			if(isNaN(trigIdVal) || trigIdVal > 27 || trigIdVal < 0) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid trigger ID '" + parameters['ID'] + "', expecting a number between 0 and 27");
				error = true;
				continue;
			}
			trigIdVal = trigIdVal.toString(2);
			trigIdVal = zeroPad(trigIdVal, 5);
			
			var trigSel = parameters['SEL'].toUpperCase();
			if(!triggerSelectMap.hasOwnProperty(trigSel)) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid trigger select '" + parameters['SEL'] + "'.");
				error = true;
				continue;
			}
			var trigSelVal = triggerSelectMap[trigSel];
			
			var trigType = parameters['TYPE'].toUpperCase();
			if(!triggerTypeMap.hasOwnProperty(trigType)) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid trigger type '" + parameters['TYPE'] + "'.  Must be LOW, HIGH, RISE, or FALL");
				error = true;
				continue;
			}
			var trigTypeVal = triggerTypeMap[trigType];
			
			var immVal = convertStringToInt(parameters['IMM']);
			if(isNaN(immVal) || immVal < 0 || immVal > 1) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid immediate value '" + parameters['IMM'] + "', must be 0 or 1");
				error = true;
				continue;
			}
			immVal = immVal.toString(2);
			
			programBin[i] = destinationVal + extVal + immVal + trigTypeVal + trigSelVal + trigIdVal + opcodeMap[command];
			
		} else if(command === "TRIG_MASK") {
			if(fields.length != 2) {
				console.log("\nInstruction: " + programTxt[i] + "\nExpected 1 argument, found " + (fields.length - 1));
				error = true;
				continue;
			}
		
			var mask = convertStringToInt(fields[1]).toString(2);
			if(isNaN(mask)) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Invalid mask value, expected a number");
				error = true;
				continue;
			} else if(mask.length > 28) {
				console.log("\nInstruction: " + programTxt[i] + "\nERROR: Mask cannot be longer than 28 bits");
				error = true;
				continue;
			} else {
				mask = zeroPad(mask, 28);
			}
			programBin[i] = mask + opcodeMap[command];
		} else if(command === "LITERAL") {
			if(fields.length === 1) {
				console.log("\nInstruction: " + programTxt[i] + "\nExpected at least 2 arguments, found 1");
				error = true;
				continue;
			}
			
			var binary = parseLiteralCommand(fields, programTxt[i]);
			
			if(binary === undefined) {
				error = true;
				continue;
			}
			
			programBin[i] = binary;
		}
		
    }
    
	var dbgString = createDebugString(programTxt, programAddrList, reverseSymbols);
	if(debug) {
		console.log(symbolsAddress);
		console.log(symbolsInstructionCount);
		console.log(programTxt);
		console.log(programBin);
		console.log(dbgString)
		
		//testProgram(programBin, expectedProgramBin, programTxt);
	}
	
	if(error) {
		return {};
	} else {
		var binaryString = formatBits(programBin);
		
		if(debug) {
			console.log(binaryString);
		}
		var results = {};
		results.asm = fileContents;
		results.dbg = dbgString;
		results.bin = binaryString;
		return results;
	}
}

var disassemble = function(binaryTxt) {
	
	var lines = binaryTxt.split(/\s+/); // split on any whitespace
	var binaryString = "";
	var binaryIndex = 0;
	var instructionList = [];
	var instructionLength = 0;
	
	for(var i=0; i<lines.length; i++) {
		binaryString = lines[i] + binaryString;
	}
	
	// Extracts the field given the bit position, and the number of bits towards the more-significant direction
	// getBinaryField(0,4) will extract bits 3:0
	var getBinaryField = function(position, size) { 
		var binaryVal = binaryString.substring(binaryString.length - position - size, binaryString.length - position);
		return parseInt(binaryVal, 2);
	}
	
	var endSeq = false;
	
	while(binaryIndex < binaryString.length) {
		var command = binaryString.substring(binaryString.length-binaryIndex-4, binaryString.length-binaryIndex);
		var instruction = {"address": "0x" + zeroPad(binaryIndex.toString(16), 4)};
			
		if(command === '0000') {
			if(endSeq) { // Ignore zeroes after sequences, this will be after the end of the pfsm sequence
				binaryIndex += 4;
				continue;
			} else {
				instruction["assembly"] = "LITERAL 4,0x0 ; Invalid instruction";
				instructionLength = 4;
			}

		} else if(command === '0001') { // REG_WRITE_IMM
			instructionLength = commandLength['REG_WRITE_IMM'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			data = zeroPad(parseInt(instructionBin.substring(0, 8), 2).toString(16).toUpperCase(), 2);
			addr = zeroPad(parseInt(instructionBin.substring(8, 18), 2).toString(16).toUpperCase(), 3);
			
			instruction["assembly"] = "REG_WRITE_IMM ADDR=0x" + addr.toString(16) + " DATA=0x" + data.toString(16);
			endSeq = false;
		} else if(command === '0010') { // REG_WRITE_MASK_IMM
			instructionLength = commandLength['REG_WRITE_MASK_IMM'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			mask = zeroPad(parseInt(instructionBin.substring(0, 8), 2).toString(16).toUpperCase(), 2);
			data = zeroPad(parseInt(instructionBin.substring(8, 16), 2).toString(16).toUpperCase(), 2);
			addr = zeroPad(disassembleAddress(instructionBin.substring(16, 26)).toString(16).toUpperCase(), 3);
			
			instruction["assembly"] =  "REG_WRITE_MASK_IMM ADDR=0x" + addr.toString(16) + " DATA=0x" + data.toString(16) + " MASK=0x" + mask;
			endSeq = false;
		} else if(command === '0011') { // REG_WRITE_VOUT_IMM
			instructionLength = commandLength['REG_WRITE_VOUT_IMM'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			delay = parseInt(instructionBin.substring(0, 6), 2);
			vout = parseInt(instructionBin.substring(6, 14), 2);
			selBin = instructionBin.substring(14, 16);
			sel = reverseVoutSelMap[selBin];
			regulatorId = parseInt(instructionBin.substring(16,20), 2);
			regulator = reverseRegulatorIdMap[regulatorId];
			
			// Convert vout binary to voltage
			var voutVolts;
			var bypass;
			if(regulator.substring(0,4) === "BUCK") {
				if(vout <= 0x0E) {
					voutVolts = vout * 0.02 + 0.3;
				} else if(vout <= 0x72) {
					voutVolts = vout * 0.005 + 0.525;
				} else if(vout <= 0xAA) {
					voutVolts = vout * 0.01 - 0.05;
				} else {
					voutVolts = vout * 0.02 - 1.76;
				}
			} else { // LDO
				if(regulator === "LDO4") {
					voutVolts = vout * 0.025 + 0.4;
					bypass = "0";
				} else {
					if((vout & 0x80) === 0x80) {
						bypass = "1";
						vout = vout & 0x7F;
					} else {
						bypass = "0";
					}
					
					voutVolts = (vout >> 1) * 0.05 + 0.4;
				}
			}
			
			voutVolts = Math.round(voutVolts * 1000) / 1000; // Round to 3 decimal places
			
			instruction["assembly"] = "REG_WRITE_VOUT_IMM DELAY=" + delay + " VOUT=" + voutVolts + "V SEL=" + sel + " REGULATOR=" + regulator;
			if(regulator.substring(0,3) === "LDO") {
				instruction["assembly"] += " BYPASS=" + bypass;
			}
			
			endSeq = false;
		} else if(command === '0100') { // REG_WRITE_VCTRL_IMM
			instructionLength = commandLength['REG_WRITE_VCTRL_IMM'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			delay = parseInt(instructionBin.substring(0, 6), 2);
			mask = zeroPad(parseInt(instructionBin.substring(6,14), 2).toString(16).toUpperCase(), 2);
			vctrl = zeroPad(parseInt(instructionBin.substring(14, 22), 2).toString(16).toUpperCase(), 2);
			delayModeBin = instructionBin.substring(22, 24);
			if (delayModeBin == '00') {
				delayMode = 'MATCH_EN';
			} else if(delayModeBin == '01') {
				delayMode = 'MATCH_ALL';
			} else if(delayModeBin == '10') {
				delayMode = 'ALWAYS';
			} else {
				console.log('Invalid delay mode "11"');
				return "";
			}
			regulatorId = parseInt(instructionBin.substring(24, 28), 2);
			regulator = reverseRegulatorIdMap[regulatorId];
			
			instruction["assembly"] = "REG_WRITE_VCTRL_IMM DELAY=" + delay + " MASK=0x" + mask + " VCTRL=0x" + vctrl + " DELAY_MODE=" + delayMode + " REGULATOR=" + regulator;
			endSeq = false;
		} else if(command === '0101') { // REG_WRITE_MASK_SREG
			instructionLength = commandLength['REG_WRITE_MASK_SREG'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			mask = zeroPad(parseInt(instructionBin.substring(0, 8), 2).toString(16).toUpperCase(), 2);
			sregId = parseInt(instructionBin.substring(8, 10), 2);
			sreg = "R" + sregId;
			addr = zeroPad(parseInt(instructionBin.substring(10, 20), 2).toString(16).toUpperCase(), 3);
			
			instruction["assembly"] = "REG_WRITE_MASK_SREG REG=" + sreg + " ADDR=0x" + addr + " MASK=0x" + mask;
			endSeq = false;
		} else if(command === '0110') { // SREG_READ_REG
			instructionLength = commandLength['SREG_READ_REG'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			sregId = parseInt(instructionBin.substring(0, 2), 2);
			sreg = "R" + sregId;
			addr = zeroPad(parseInt(instructionBin.substring(2,12), 2).toString(16).toUpperCase(), 3);
			
			instruction["assembly"] = "SREG_READ_REG REG=" + sreg + " ADDR=0x" + addr;
			endSeq = false;
		} else if(command === '0111') { // WAIT
			instructionLength = commandLength['WAIT'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			skipCount = parseInt(instructionBin.substring(0, 6), 2);
			timeout = parseInt(instructionBin.substring(6,12), 2);
			conditionTypeBin = instructionBin.substring(12,14);
			conditionType = reverseTriggerTypeMap[conditionTypeBin];
			conditionSelBin = instructionBin.substring(14,20);
			conditionSel = reverseConditionSelectMap[conditionSelBin];
			
			instruction["assembly"] = "WAIT SKIP_COUNT=" + skipCount + " TIMEOUT=" + timeout + " COND=" + conditionSel + " TYPE=" + conditionType;
			endSeq = false;
		} else if(command === '1000') { // DELAY_IMM
			instructionLength = commandLength['DELAY_IMM'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			delay = parseInt(instructionBin.substring(0,8), 2);
			
			instruction["assembly"] = "DELAY_IMM " + delay;
			endSeq = false;
		} else if(command === '1001') { // DELAY_SREG
			instructionLength = commandLength['DELAY_SREG'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			sreg = "R" + parseInt(instructionBin.substring(0,2), 2);
			instruction["assembly"] = "DELAY_SREG " + sreg;
			endSeq = false;
		} else if(command === '1010') { // TRIG_SET
			instructionLength = commandLength['TRIG_SET'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			addr = parseInt(instructionBin.substring(0,13), 2);
			ext = instructionBin.substring(13,14);
			imm = instructionBin.substring(14,15);
			trigType = reverseTriggerTypeMap[instructionBin.substring(15,17)];
			trigSel = reverseTriggerSelectMap[instructionBin.substring(17,23)];
			trigId = parseInt(instructionBin.substring(23,28), 2);
			if(ext === "1") {
				addr = reverseExternalStateMap[addr];
			} else {
				addr = "0x" + zeroPad(addr.toString(16), 4);
			}
			
			instruction["assembly"] = "TRIG_SET DEST=" + addr + " ID=" + trigId + " SEL=" + trigSel + " TYPE=" + trigType + " IMM=" + imm + " EXT=" + ext;
			endSeq = false;
		} else if(command === '1011') { // TRIG_MASK
			instructionLength = commandLength['TRIG_MASK'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			trigMask = zeroPad(parseInt(instructionBin.substring(0, 28), 2).toString(16).toUpperCase(), 7);
			
			instruction["assembly"] = "TRIG_MASK 0x" + trigMask;
			endSeq = false;
		} else if(command === '1100') { // END
			instruction["assembly"] = "END";
			instructionLength = commandLength['END'];
			var instructionBin = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
			endSeq = true;
		} else if(command === '1101') {
			instruction["assembly"] = "LITERAL 4,0xD ; Invalid instruction";
			conole.log("invalid instruction, address " + binaryIndex);
			instructionLength = 4;
			endSeq = false;
		} else if(command === '1110') {
			instruction["assembly"] = "LITERAL 4,0xE ; Invalid instruction";
			conole.log("invalid instruction, address " + binaryIndex);
			instructionLength = 4;
			endSeq = false;
		} else if(command === '1111') {
			instruction["assembly"] = "LITERAL 4,0xF ; Invalid instruction";
			conole.log("invalid instruction, address " + binaryIndex);
			instructionLength = 4;
			endSeq = false;
		} else {
			if(command !== '0' && command !== '00' && command !== '000') { // Only throw an error if we are not at the end of the sequence, where we would see a shorter command of zeroes
				console.log("Unexpected command '" + command + "', expecting 4-bit binary command");
				return "";
			} else {	// At end of sequence
				break;
			}
		}
		
		instruction["binary"] = binaryString.substring(binaryString.length-binaryIndex-instructionLength, binaryString.length-binaryIndex);
		instructionList.push(instruction);
		binaryIndex += instructionLength;
	}
	
	return instructionList;
}

var parseLiteralCommand = function(fields, instructionTxt) {
	var binary = "";
	
	for(j=1; j<fields.length; j++) {
		
		var argument = "";
		
		if(fields[j].indexOf("=") === -1) {
			argument = fields[j];
		} else {
			argument = fields[j].split("=")[1];
		}
		
		if(argument.indexOf(",") === -1) {
			console.log("\nInstruction: " + instructionTxt + "\nField '" + fields[j] + "', is missing the number of bits.  The format should be '<#bits>,<value>'");
			return undefined;
		}
		
		var argValue = argument.split(",");
		var numBits = convertStringToInt(argValue[0]);
		var value = convertStringToInt(argValue[1]);
		
		if(isNaN(numBits)) {
			console.log("\nInstruction: " + instructionTxt + "\nInvalid number of bits: " + argValue[0]);
			return undefined;
		}
		
		if(isNaN(value)) {
			console.log("\nInstruction: " + instructionTxt + "\nInvalid value: " + argValue[1]);
			return undefined;				
		}
		
		if(value.toString(2).length > numBits) {
			console.log("\nInstruction: " + instructionTxt + "\nInvalid value: " + argValue[1] + ", it is longer than the specified bit size");
			return undefined;
		}
		
		binary = zeroPad(value.toString(2), numBits) + binary;
	}
	
	return binary;
}

// Parses the provided fields for each parameter given in the paramList argument.
// It will parse by name if provided, otherwise it will parse by position, in the same order of paramList.
// Optional parameters contain the string " (optional)" after the parameter name, eg. "BYPASS (optional)".
// Returns false on error, otherwise returns an object mapping the parameter name to the value.
var parseParameters = function(paramList, fields, instruction) {
	var parameters = {};
	var error = false;
	var optional = [];
	
	// Check if the parameter is optional
	for(var i=0; i<paramList.length; i++) {
		if(paramList[i].indexOf(" (optional)") !== -1) {
			paramList[i].replace(" (optional)", "");
			optional.push(true);
		} else {
			optional.push(false);
		}
	}
	
	var maxArgs = optional.length;
	var minArgs = optional.length - optional.filter(Boolean).length;	// optional.filter(Boolean).length returns number of true values in array
	
	if((fields.length-1) < minArgs || (fields.length-1) > maxArgs) {
		if(minArgs === maxArgs) {
			console.log("\nInstruction: " + instruction + "\nExpected " + paramList.length + " arguments, found " + (fields.length - 1));
		} else {
			console.log("\nInstruction: " + instruction + "\nExpected " + minArgs + " to " + maxArgs + " arguments, found " + (fields.length - 1));
		}
		
		return false;	// Return false on error, because it's easier to check for than an empty object		
	}
	
	for(var i=0; i<paramList.length; i++) {
		
		
		// Parse the parameter
		if(optional[i] && fields[i+1] === undefined) {
			continue;
		} else {
			if(fields[i+1].indexOf("=") !== -1) {
				var parsedParameters = fields[i+1].split("=");
				parameters[parsedParameters[0].toUpperCase()] = parsedParameters[1];
			} else {
				parameters[paramList[i]] = fields[i+1];
			}
		}
	}
	
	for(i=0; i<paramList.length; i++) {
		if(!optional[i] && parameters[paramList[i]] === undefined) {
			console.log("\nInstruction: " + instruction + "\n" + paramList[i] + " argument not found.  It must be argument #" + (i+1) + ", or else add '" + paramList[i] + "='");
			error = true;
			continue;
		}
	}
	
	if(error === true) {
		return false;	// Return false on error, because it's easier to check for than an empty object
	} else {
		return parameters;
	}
}

// Converts the given time (including unit) in to nanoseconds
var convertTimeToNs = function(timeStr) {
	if(timeStr.substring(timeStr.length-2).toLowerCase() == "ns") {
		time = convertStringToInt(timeStr.substring(0, timeStr.length-2));
		if(isNaN(time)) {
			return -1;
		} else {
			return time;
		}
	}
	
	if(timeStr.substring(timeStr.length-2).toLowerCase() == "us") {
		time = convertStringToInt(timeStr.substring(0, timeStr.length-2));
		if(isNaN(time)) {
			return -1;
		} else {
			return time * 1e3;
		}
	}
	
	if(timeStr.substring(timeStr.length-2).toLowerCase() == "ms") {
		time = convertStringToInt(timeStr.substring(0, timeStr.length-2));
		if(isNaN(time)) {
			return -1;
		} else {
			return time * 1e6;
		}
	}
	
	if(timeStr.substring(timeStr.length-1).toLowerCase() == "s") {
		time = convertStringToInt(timeStr.substring(0, timeStr.length-1));
		if(isNaN(time)) {
			return -1;
		} else {
			return time * 1e9;
		}
	}
	
	return -1;
}

// Converts number of nanoseconds in to a time including units
var convertNsToTime = function(time) {
	if(time < 1e3) {
		return time + "ns";
	}
	if(time < 1e6) {
		return +((time/1e3).toFixed(3)) + "us"; // This rounds to 3 digits, but drops trailing 0s
	}
	if(time < 1e9) {
		return +((time/1e6).toFixed(3)) + "ms";
	}
	return +((time/1e9).toFixed(3)) + "s";
}

// Converts a delay time in to a delay value, given a certain delay step.  It will round to the nearest delay value.
// numBits is the number of bits for the delay field.  Some instructions have 6 bits, some have 8 bits.
// For example, delayTime=5ms, delayStep=1.6ms, it will return 3.
// Returns -1 on error
var convertDelayTimeToValue = function(delayTime, delayStep, instruction, numBits) {
	
	// If it's a number already, just return the number
	if(!isNaN(delayTime))  {
		return parseInt(delayTime);
	}
	delayTimeNs = convertTimeToNs(delayTime);
	if(delayStep === -1 && delayTimeNs !== 0) {
		console.log("\nInstruction: " + instruction + "\nERROR: No delay step is set.  Define default PFSM_DELAY_STEP value, or use SET_DELAY command before this instruction");
		return -1;
	}
	if(delayTimeNs === -1) {
		console.log("\nInstruction: " + instruction + "\nERROR: Invalid delay time '" + delayTime + "', expecting a number with units of ns, us, ms, or s");
		return -1;
	}
	if(delayTimeNs < 0) {
		console.log("\nInstruction: " + instruction + "\nERROR: Invalid delay time '" + delayTime + "', must not have negative delay");
		return -1;
	}
	if(delayTimeNs > (Math.pow(2,numBits)-1)*107.4e9) {
		console.log("\nInstruction: " + instruction + "\nERROR: Invalid delay time '" + delayTime + "', maximum delay is " + ((Math.pow(2, numBits)-1) * 107.4) + " seconds");
		return -1;
	}
	var delayVal = Math.round(delayTimeNs / delayStep);
	if(delayVal === 0 && delayTimeNs !== 0) {
		console.log("\nInstruction: " + instruction + "\nWARNING: Delay time of " + delayTime + " is rounded to 0, with current delay step of " + convertNsToTime(delayStep));
	} else if(delayVal > Math.pow(2, numBits)-1) {
		console.log("\nInstruction: " + instruction + "\nERROR: Delay time of " + delayTime + " cannot be achieved with current step size of " + convertNsToTime(delayStep));
		console.log("\nUse SET_DELAY command to choose a larger delay step size.");
		return -1;
	}
	
	return delayVal;
}

// Takes a list of instructions, and returns back a string in 32-bit lines
var formatBits = function(instructions) {
	var binaryArray = [];
	var currLine = "";
	
	for(var i=0; i<instructions.length; i++) {
		var instruction = instructions[i];
		if(currLine.length + instruction.length <= 32) {
			currLine = instruction + currLine;
		} else {
			var bitsRemaining = 32-currLine.length;
			currLine = instruction.substring(instruction.length - bitsRemaining) + currLine;
			binaryArray.push(currLine);
			currLine = instruction.substring(0, instruction.length-bitsRemaining);
		}
	}
	
	if(currLine.length < 32) {
		currLine = zeroPad(currLine, 32);
	}
	binaryArray.push(currLine);
	
	var binaryString = "";
	for(i=0; i<binaryArray.length; i++) {
		binaryString += binaryArray[i] + "\n";
	}
	
	return binaryString;
}

var createDebugString = function(instructions, instructionAddrs, reverseSymbols) {
	var debugString = ""
	for(var i=0; i<instructions.length; i++) {
		var address = instructionAddrs[i];
		if(reverseSymbols.hasOwnProperty(address)) {
			debugString += reverseSymbols[address] + ":\n";
		}
		debugString += "0x" + instructionAddrs[i].toString(16) + ": " + instructions[i] + "\n";
	}
	return debugString;
}

var testProgram = function(assembledProgram, expectedProgram, programTxt) {
	var pass = true;
	for(var i=0; i<assembledProgram.length; i++) {
		if(assembledProgram[i] != expectedProgram[i]) {
			console.log("TEST FAIL! Instruction: " + programTxt[i] + "\nAssembled value: " + assembledProgram[i] + "\nExpected value:  " + expectedProgram[i]);
			pass = false;
		}
	}
	
	if(pass) {
		console.log("TEST PASS!");
	}
}

// Adds leading zeroes to a string, to get it to the provided length.
// Example zeroPad("11", 8) = "00000011"
var zeroPad = function(str, length) {
	
	padding = "";
	for(var i=0; i<length-str.length; i++) {
		padding += "0";
	}
	return padding + str;
}

// Does error checking, and calculate address mapping from input address to PFSM address.
// Page 4 (WD) is converted to 0x3XX, and page 3 (SRAM) is unwritable and throws an error.
// Returns -1 on error.
var calculateAddress = function(address, instruction) {
	if(address === "NaN") {
		console.log("\nInstruction: " + instruction + "\nInvalid register address, expecting a number");
		return -1;
	}
	
	if(address.length == 10 && address.substring(0,2) == "11") {
		console.log("\nInstruction: " + instruction + "\nCannot write to page 3 (SRAM) of address space.  Must be page 0, 1, 2, or 4");
		return -1;
	}
	
	// Maps WD page 4 in to PFSM page 3
	if(address.length == 11 && address.substring(0, 3) == "100") {
		address = "11" + address.substring(3);
	}
	
	if(address.length > 10) {
		console.log("\nInstruction: " + instruction + "\nInvalid register address, must be 10 bits or less. 11 bits is allowed only if writing to page 4");
		return -1;
	}
	
	return zeroPad(address, 10);
}

// Calculates address mapping from PFSM address to input address.
// The watchdog page is assembled as 0x3XX, but is converted in assembly to 0x4XX
// Returns -1 on error.
var disassembleAddress = function(address) {
	address = parseInt(address, 2);
	if(address >= 0x300) {
		address += 0x100;
	}
	return address;
}

// Our jjs engine (to run from command line) doesn't support Number() conversion leading with 0b.
// So this will use parseInt if a binary value is proveded.
var convertStringToInt = function(str) {
	if(str.substring(0, 2) == "0b") {
		return parseInt(str.substring(2), 2);
	}
	else {
		return Number(str);
	}
}
